package decatur.de.rigorous.numutil;

import decatur.de.rigorous.parse.Expression;

/**
   This class defines a root finder for a scalar expression in 1 unknown.
   @author Wolfgang Kuehn
   @version 1.0
   @see parse.Expression
*/
public class RootFinder {
    String message = "\nRoot not found!";

    /**
       Find enclosure of root.
       @param x0 Hint where to start to search.
       @param expression Find root of this expression in 1 unknown.
       @return An enclosure of the root.
       @exception ArithmeticException If root cannot be found.
    */ 
    public Interval find (double x0, Expression expression)
	throws ArithmeticException {

	double rootinf = newton (x0, -1, expression);
	double rootsup = newton (x0, 1, expression);
	
	if ( rootinf>=rootsup ) 
	    throw new ArithmeticException (message);
	
	Interval x = new Interval (rootinf, rootsup);
	expression.eval (x);
	
	if ( expression.derivative().contains (0.0) )
	    throw new ArithmeticException (message);
	
	return x;
    }
    
    /**
       Find possible lower or upper bound for root.
       @param x0 Hint where to start to search.
       @param dir Find lower bound if -1, find upper bound if 1.
       @param expression Find root of this expression.
    */

    protected double newton (double x0, int dir, Expression expression) 
	throws ArithmeticException {

	double x, x_old, y, dy;
	int k;
	expression.eval(x0);
	dy = expression.derivative().mid();
	int np;			// np = 1  if positive slope
				// np = -1 if negative slope
	if ( dy>0.0 ) np = 1;
	else np = -1;
	
	for (x=x0, k=0; k<100; k++) {
	    expression.eval(x);
	    if ( np*dir==-1 ) {
		y = expression.value().sup;
		dy = expression.derivative().sup;
	    } else {
		y = expression.value().inf;
		dy = expression.derivative().sup;
	    }
	    x_old = x;
	    x -= y/dy;
	    //	System.out.println ("newton: "+k+" "+x+" "+x_old);
	    if ( Math.abs(x_old-x)/(Math.abs(x)+1e-14)<1e-14 ) {
		for (int l=0; l<20; l++) {
		    //	    System.out.println ("eps:" + x);
		    if ( dir*np*y>=0.0 ) return x;
		    x += (Math.abs(x)+1.0)*dir*1e-14;
		    expression.eval(x);
		    if ( np*dir==-1 )
			y = expression.value().sup;
		    else
			y = expression.value().inf;
		}
		break;
	    }
	}
	throw new ArithmeticException (message);
    }
}
