package decatur.de.rigorous.parse;

import decatur.de.rigorous.numutil.Interval;

/**
 * This class defines a scalar expression in N unknowns.
 * 
 * @author Wolfgang Kuehn
 * @version 1.0
 */

public class Expression {
	private int tempcount, maxadr = 0;
	private int code[];
	private Interval data[];
	private int nvars;
	private Gradient gradient;
	private String token, stream;
	private int tokencount;
	private char symc;
	private int type;

	private final int INTEGER = 1;
	private final int DOUBLE = 2;
	private final int WORD = 3;
	private final int EOF = 4;
	private final int PUNCT = 5;

	private final int ADD = 1;
	private final int SUB = 2;
	private final int MUL = 3;
	private final int DIV = 4;
	private final int NEG = 5;
	private final int POW = 6;
	private final int SQRT = 7;
	private final int SIN = 8;
	private final int COS = 9;
	private final int EXP = 10;
	private final String functoken[] = { "sqrt", "sin", "cos", "exp" };
	private int ZERO, ONE;
	private String[] names;

	/**
	 * Constructs an Expression object.
	 * 
	 * @param s
	 *            The expression.
	 * @param names
	 *            The names of all N variables.
	 * @exception Exception
	 *                On syntax error.
	 */
	public Expression(String s, String[] names) throws Exception {
		code = new int[1000];
		data = Interval.vector(1000);
		if (names != null)
			nvars = names.length;
		else
			nvars = 0;
		tempcount = nvars;
		this.names = names;
		ZERO = newtemp();
		data[ZERO].set(Interval.ZERO);
		ONE = newtemp();
		data[ONE].set(Interval.ONE);
		stream = s;
		tokencount = -1;
		nextToken();
		gradient = expr();
		// dump ();
		if (type != EOF)
			throw new Exception("End of expression expected");
	}

	/**
	 * Evaluate expression.
	 * 
	 * @param vars
	 *            Interval array for values of all N variables.
	 */
	public void eval(Interval[] vars) {
		if (vars != null)
			for (int i = 0; i < nvars; i++)
				data[i].set(vars[i]);

		evaluate();
	}

	/**
	 * Evaluate expression with single Interval argument (in case N=1).
	 * 
	 * @param var
	 *            Interval value of single variable.
	 */
	public void eval(Interval var) {
		data[0].set(var);
		evaluate();
	}

	/**
	 * Evaluate expression with single double argument (in case N=1).
	 * 
	 * @param var
	 *            double value of single variable.
	 */
	public void eval(double var) {
		data[0].set(var);
		evaluate();
	}

	/**
	 * @return The value of the expression resulting from the previous
	 *         evaluation.
	 */
	public Interval value() {
		return data[gradient.e];
	}

	/**
	 * @return The partial derivative of the expression with respect to the i-th
	 *         variable resulting from the previous evaluation.
	 */
	public Interval partial(int i) {
		return data[gradient.de[i]];
	}

	/**
	 * @return The derivative of the expression (in case N=1) resulting from the
	 *         previous evaluation.
	 */
	public Interval derivative() {
		return partial(0);
	}

	private void next() {
		tokencount++;
		if (tokencount == stream.length())
			symc = '@';
		else
			symc = stream.charAt(tokencount);
		// System.out.println("symc "+symc+" "+tokencount);
	}

	private void unnext() {
		symc = stream.charAt(--tokencount);
	}

	private void nextToken() throws Exception {
		do
			next();
		while (Character.isWhitespace(symc));
		int start = tokencount;
		// System.out.println("nextToken: "+stream.substring(start)+" "+symc);

		// JDK1.1: Character.isJavaIdentifierStart(symc)
		if (Character.isLowerCase(symc) || Character.isUpperCase(symc)) {/*
																		 * Read
																		 * an
																		 * identifier
																		 * .
																		 */
			next();
			while (Character.isLowerCase(symc) || Character.isUpperCase(symc))
				next();
			// System.out.println("start "+start+" "+tokencount);
			token = stream.substring(start, tokencount);
			type = WORD;
			unnext();
		} else if (Character.isDigit(symc) || symc == '.') {
			type = INTEGER;
			while (Character.isDigit(symc))
				next();

			if (symc == '.') {
				next();
				while (Character.isDigit(symc))
					next();
				type = DOUBLE;
			}

			if (symc == 'e' || symc == 'E') {
				next();
				int estart = tokencount;
				if (symc == '-' || symc == '+') {
					estart++;
					next();
				}

				while (Character.isDigit(symc))
					next();
				if (tokencount == estart)
					throw new Exception("No digits in exponent");

				type = DOUBLE;
			}
			token = stream.substring(start, tokencount);
			unnext();
		} else if (symc == '@') {
			type = EOF;
		} else {
			type = PUNCT;
			token = "op";
		}

		// System.out.println("token: "+token+" "+symc+ " "+type);
	}

	private void evaluate() {
		int adr = 0;
		while (adr < maxadr) {
			int op = code[adr++];
			Interval n = data[code[adr++]];
			Interval l = data[code[adr++]];
			n.set(l);
			switch (op) {
			case ADD:
				n.add(data[code[adr++]]);
				break;
			case SUB:
				n.sub(data[code[adr++]]);
				break;
			case MUL:
				n.mul(data[code[adr++]]);
				break;
			case DIV:
				n.div(data[code[adr++]]);
				break;
			case NEG:
				n.neg();
				break;
			case POW:
				n.pow(code[adr++]);
				break;
			case SQRT:
				n.sqrt();
				break;
			case SIN:
				n.sin();
				break;
			case COS:
				n.cos();
				break;
			case EXP:
				n.exp();
				break;
			default: // Never happens
				// System.out.println("Bad operand");
				break;
			}
		}
	}

	private int newtemp() {
		return tempcount++;
	}

	private int add(int op, int l, int r) {
		code[maxadr++] = op;
		int n = code[maxadr++] = newtemp();
		code[maxadr++] = l;
		code[maxadr++] = r;
		return n;
	}

	private int add1(int op, int l) {
		code[maxadr++] = op;
		int n = code[maxadr++] = newtemp();
		code[maxadr++] = l;
		return n;
	}

	private int optimize(int op, int l, int r) {
		switch (op) {
		case ADD:
		case SUB:
			if (l == ZERO && r == ZERO)
				return ZERO;
			if (r == ZERO)
				return l;
			if (l == ZERO) {
				if (op == ADD)
					return r;
				else
					return add1(NEG, r);
			}
			break;
		case MUL:
			if (l == ZERO || r == ZERO)
				return ZERO;
			if (l == ONE)
				return r;
			if (r == ONE)
				return l;
			break;
		default:
			break;
		}
		return add(op, l, r);
	}

	private Gradient generate(int op, Gradient l, Gradient r) {
		Gradient result = new Gradient(nvars);
		int d = 0;
		result.e = optimize(op, l.e, r.e);
		if (op == DIV)
			d = optimize(POW, r.e, 2);

		for (int i = 0; i < nvars; i++) {
			int s1, s2;
			switch (op) {
			case ADD:
			case SUB:
				result.de[i] = optimize(op, l.de[i], r.de[i]);
				break;
			case MUL:
				s1 = optimize(MUL, l.e, r.de[i]);
				s2 = optimize(MUL, l.de[i], r.e);
				result.de[i] = optimize(ADD, s1, s2);
				break;
			case DIV:
				s1 = optimize(MUL, l.e, r.de[i]);
				s2 = optimize(MUL, l.de[i], r.e);
				s1 = optimize(SUB, s2, s1);
				result.de[i] = optimize(DIV, s1, d);
				break;
			default:
				break;
			}
		}
		return result;
	}

	private Gradient generateNEG(Gradient l) {
		Gradient result = new Gradient(nvars);
		result.e = add1(NEG, l.e);

		for (int i = 0; i < nvars; i++)
			result.de[i] = add1(NEG, l.de[i]);
		return result;
	}

	private Gradient generatePOW(Gradient l, int exp) {
		Gradient result = new Gradient(nvars);
		result.e = optimize(POW, l.e, exp);
		int p = -1;
		for (int i = 0; i < nvars; i++) {
			if (l.de[i] != ZERO && p == -1) {
				p = optimize(POW, l.e, exp - 1);
				int iexp = newtemp();
				data[iexp].set(exp);
				p = optimize(MUL, p, iexp);
			}
			result.de[i] = optimize(MUL, p, l.de[i]);
		}
		return result;
	}

	private Gradient generateFunc(int op, Gradient l) {
		Gradient result = new Gradient(nvars);
		result.e = add1(op, l.e);
		int f = -1;
		for (int i = 0; i < nvars; i++) {
			if (l.de[i] != ZERO && f == -1) {
				switch (op) {
				case SQRT:
					int onehalf = newtemp();
					data[onehalf].set(0.5);
					f = optimize(DIV, onehalf, result.e);
					break;
				case SIN:
					f = add1(COS, l.e);
					break;
				case COS:
					f = add1(SIN, l.e);
					f = add1(NEG, f);
					break;
				case EXP:
					f = result.e;
					break;
				default:
					break;
				}
			}
			result.de[i] = optimize(MUL, f, l.de[i]);
		}
		return result;
	}

	private boolean checkfor(char c) {
		return (symc == c);
	}

	private void ismissing(char c) throws Exception {
		if (!checkfor(c))
			throw new Exception("Character " + "'" + c + "'" + " expected");
	}

	private double scanNumber() throws Exception {
		int sign = 1;
		if (checkfor('-')) {
			sign = -1;
			nextToken();
		} else if (checkfor('+'))
			nextToken();
		if (!(type == INTEGER || type == DOUBLE))
			throw new Exception("Number expected");
		return sign * Double.valueOf(token).doubleValue();
	}

	private Gradient scanInterval() throws Exception {
		Gradient result = new Gradient(nvars);
		for (int i = 0; i < nvars; i++) {
			result.de[i] = ZERO;
		}
		result.e = newtemp();
		Interval w = data[result.e];
		if (checkfor('[')) {
			nextToken();
			w.inf = scanNumber();
			nextToken();
			ismissing(',');
			nextToken();
			w.sup = scanNumber();
			nextToken();
			ismissing(']');
			if (w.inf > w.sup)
				throw new Exception("infimum > supremum");
		} else
			w.set(scanNumber());
		nextToken();
		return result;
	}

	private Gradient ident() throws Exception {
		Gradient result = new Gradient(nvars);
		int index;
		for (index = 0; index < nvars; index++)
			if (token.equals(names[index]))
				break;
		if (index == nvars)
			throw new Exception("Variable " + token + " undefined");
		result.e = index;
		for (int i = 0; i < nvars; i++) {
			result.de[i] = (i == index) ? ONE : ZERO;
		}
		return result;
	}

	private Gradient atom() throws Exception {
		Gradient result;
		if (checkfor('[') || type == DOUBLE || type == INTEGER)
			result = scanInterval();
		else if (type == WORD) {
			int op;
			for (op = SQRT; op <= EXP; op++)
				if (token.equals(functoken[op - SQRT]))
					break;
			if (op > EXP)
				result = ident();
			else {
				nextToken();
				ismissing('(');
				nextToken();
				result = expr();
				ismissing(')');
				result = generateFunc(op, result);
			}
			nextToken();
		} else
			throw new Exception("Number or variable expected");

		return result;
	}

	private Gradient expr() throws Exception {
		Gradient l;
		if (checkfor('-')) {
			nextToken();
			l = factor();
			l = generateNEG(l);
		} else {
			if (checkfor('+'))
				nextToken();
			l = factor();
		}

		while (checkfor('+') || checkfor('-')) {
			int op = checkfor('+') ? ADD : SUB;
			nextToken();
			l = generate(op, l, factor());
		}
		return l;
	}

	private Gradient factor() throws Exception {
		Gradient l = term();
		while (checkfor('*') || checkfor('/')) {
			int op = checkfor('*') ? MUL : DIV;
			nextToken();
			l = generate(op, l, term());
		}
		return l;
	}

	private Gradient term() throws Exception {
		Gradient result = new Gradient(nvars);
		if (checkfor('(')) {
			nextToken();
			result = expr();
			ismissing(')');
			nextToken();
		} else
			result = atom();

		if (checkfor('^')) {
			nextToken();
			double d = scanNumber();
			if (type != INTEGER)
				throw new Exception("Integer expected");
			result = generatePOW(result, (int) d);
			nextToken();
		}
		return result;
	}
}
